clear, close all, clc;
% Programmer   : Muhammad Agus Zainuddin
% Last editted : 27 Oct 2014
% Description  : Nanonetworks Minimum Energy (NME) Coding for image
%                transmission

%-----------   Input   ----------------
% lena256.bmp
% cancer256.bmp
% blood1.tif

%-----------  Input  ----------------
I = double(imread('lena256.bmp'));          % Read the Image
data = reshape(I',1,size(I,1)*size(I,2));   % convert to single stream symbols
data_bit = de2bi(data,8,'left-msb');        % convert each symbol to binary
clear data;
binary_seq = reshape(data_bit',1,size(data_bit,1)*size(data_bit,2)); 
clear data_bit;
Num_Uncode = sum(binary_seq);                       % Number of 1 in binary input stream
Percent_uncode = 100 * Num_Uncode / length(binary_seq);   % Percentage of 1 in binary input stream

% --- Bit Error Rate (BER)---
BER = 1e-1;
% --- Generate the noise ---
noise = zeros(size(binary_seq));    % Noise initialization
temp  = noise ;                     % Temporer variable
p1 = BER;       % Probability receiving 0 when transmitting 1
p2 = BER/25;    % Probability receiving 1 when transmitting 0
for j = 1:length(binary_seq),
    if binary_seq(j) == 1,  % Conver input 1 if there is an error
        acak = rand;
        if acak < p1,
            noise(j) = 1;
        end
    else
        if acak < p2,       % Conver input 0 if there is an error
            noise(j) = 1;
        end
    end
end
clear temp;
% --- Received Uncode Image ---
binary_recv = mod(binary_seq + noise,2);    % Add signal with noise
clear noise;
symbol_recv = bi2de(buffer(binary_recv,8)','left-msb');
clear binary_recv;
Image_recv = buffer(symbol_recv,256)';
clear symbol_recv;
%Image_recont = uint8(Image_recv);
Q = 255;
MSE_uncode  = sum(sum((Image_recv - I).^ 2)) / (256^2);
Image_recv = uint8(Image_recv);
PSNR_uncode = 10*log10(Q*Q/MSE_uncode); 

% --- NME encoding n-bits ---
n_bits = [2 3 4 8 16];%
for i = 1:length(n_bits),  
    disp(['NME1 ', num2str(n_bits(i)),' bits']);
    % Convert to n bits
    binary_nBits = [];
    binary_nBits = buffer(binary_seq,n_bits(i))'; symbol_nBits = [];
    symbols = bi2de(binary_nBits,'left-msb')';
    clear binary_nBits; simbol = [];
    simbol = unique(symbols); freq = [];
    % Statistic of symbols
    sta = zeros(size(simbol));
    for j = 1:length(simbol),
        sta(j) =  length(find(symbols==simbol(j)));
    end
    freq = zeros(length(simbol),2);
    for j = 1:length(simbol),
        freq(j,:) = [simbol(j) sta(j)];
    end
    % clear simbol; 
    clear sta;    % Ordered symbols by its occurance
    occur = zeros(size(freq));
    for j = 1:size(freq,1) - 1, 
        [t1,t2] = max(freq(:,2));   % t2 = row of max occurance
        occur(j,:) = freq(t2,:);
        freq(t2,:) = [];
    end
    occur(end,:) = freq(end,:); clear freq;
    if (n_bits(i) == 2),
        N_sym = [0 1 2 3];
    elseif (n_bits(i) == 3),
        N_sym = [0 1 2 4 3 5 6 7];
    elseif (n_bits(i)==4),
        N_sym = [0 2 4 1 8 5 9 10 6 3 12 7 11 13 14 15];
    elseif (n_bits(i)==8),
        load comb_nme_8.mat;
        N_sym = nme8; clear nme8;
    elseif (n_bits(i)==16),
        load comb_nme_16.mat;
        N_sym = nme16; clear nme16;
%     elseif (n_bits(i)==24),
%         load comb_nme_24.mat;
%         N_sym = s24; clear s24;
    end
    M_sym = N_sym(1:length(simbol)); %clear N_sym;
    NME_sym = de2bi(M_sym,n_bits(i),'left-msb');
    
    % NME Output
    trans_bits = zeros(length(symbols),n_bits(i));
    for j = 1:length(symbols),
        ind = find(symbols(j) == occur(:,1));    %ind2 indeks coding
        trans_bits(j,:) = NME_sym(ind,:);
    end
    clear NME_sym;
    transmitted_bit = reshape(trans_bits',1,size(trans_bits,1)*size(trans_bits,2));
    clear trans_bits;
    % Transmission Process
    noise_NME = zeros(size(transmitted_bit));
    for j = 1:length(noise_NME),
        acak = rand;
        if transmitted_bit(j) == 1,  % Conver input 1 if there is an error            
            if acak < p1,
                noise_NME(j) = 1;
            end
        else
            if acak < p2,           % Conver input 0 if there is an error
                noise_NME(j) = 1;
            end
        end
    end
    recv_bits = mod(transmitted_bit + noise_NME,2);
%     recv_bits = transmitted_bit;
    clear transmitted_bit; clear noise_NME;
    symbol_recv = bi2de(buffer(recv_bits, n_bits(i))','left-msb');    
    % Decoding nbits
    disp('Decoding time'),
    symbol_est = zeros(size(symbol_recv));
    tic
    for j = 1:length(symbol_recv),        
        ind = find(symbol_recv(j) == N_sym);
        if ind > length(simbol),
            symbol_est(j) = 0;
        else
            symbol_est(j) = occur(ind,1);
        end
    end
    toc
    clear symbol_recv;
    sym_bit = de2bi(symbol_est,n_bits(i),'left-msb'); clear symbol_est;
    recv_bit = reshape(sym_bit',1,size(sym_bit,1)*size(sym_bit,2));
    clear sym_bit;
    NSC4_recv = bi2de(buffer(recv_bit,8)','left-msb');           
    Image_est = buffer(NSC4_recv,256)'; clear NSC4_recv;
    Q = 255;
    if (n_bits(i) == 2),
        MSE  = sum(sum((Image_est - I).^ 2)) / (256^2);
        Image_2 = uint8(Image_est);
        PSNR_2bits = 10*log10(Q*Q/MSE);
    elseif (n_bits(i) == 3),
        Image_est = Image_est(1:256,:);
        MSE  = sum(sum((Image_est - I).^ 2)) / (256^2);
        Image_3 = uint8(Image_est);
        PSNR_3bits = 10*log10(Q*Q/MSE);
    elseif (n_bits(i) == 4),
        Image_est = Image_est(1:256,:);
        MSE  = sum(sum((Image_est - I).^ 2)) / (256^2);
        Image_4 = uint8(Image_est);
        PSNR_4bits = 10*log10(Q*Q/MSE);
    elseif (n_bits(i) == 8),
        Image_est = Image_est(1:256,:);
        MSE  = sum(sum((Image_est - I).^ 2)) / (256^2);
        Image_8 = uint8(Image_est);
        PSNR_8bits = 10*log10(Q*Q/MSE);
    elseif (n_bits(i) == 16),
        Image_est = Image_est(1:256,:);
        MSE  = sum(sum((Image_est - I).^ 2)) / (256^2);
        Image_16 = uint8(Image_est);
        PSNR_16bits = 10*log10(Q*Q/MSE);
    end
end

subplot 241, imshow(uint8(I)), title('Original'),
subplot 242, imshow(Image_recv), title('Original'),
   xlabel(['PSNR = ',num2str(PSNR_uncode),' dB'])   
subplot 243, imshow(Image_2), title('NME 2 bits'),
   xlabel(['PSNR = ',num2str(PSNR_2bits),' dB'])   
subplot 244, imshow(Image_3), title('NME 3 bits'),
   xlabel(['PSNR = ',num2str(PSNR_3bits),' dB'])   
subplot 245, imshow(Image_4), title('NME 4 bits'),
   xlabel(['PSNR = ',num2str(PSNR_8bits),' dB']) 
subplot 246, imshow(Image_8), title('NME 8 bits'),
   xlabel(['PSNR = ',num2str(PSNR_8bits),' dB']) 
subplot 247, imshow(Image_16), title('NME1 16 bits'),
   xlabel(['PSNR = ',num2str(PSNR_16bits),' dB']) 
% subplot 236, imshow(Image_20), title('NME1 20 bits'),
%    xlabel(['PSNR = ',num2str(PSNR_20bits),' dB'])     

